/*
 * Skeleton Intersections plugin
 * 
 * Author: Vilppu Tuominen (vilppu.tuominen@uta.fi)
 * Website: http://jvsmicroscope.uta.fi/software/
 * Updated: 24.5.2010
 *
 * Description:
 *
 * 1. Take an 8-bit binary input image, which is assumed to be skeletonized
 *    (e.g., by using the built-in ImageJ tool Binary / Skeletonize).
 * 2. Detect the intersecting pixels that have two or more connecting
 *    branches.
 * 3. Mark the detected intersections with a single pixel.
 * 4. Display an additional pseudo-colored result image showing an overlay
 *    of the intersection locations using red color.
 * 5. Generates a row to the Results table containing the number of detected
 *    intersections.
 *  
 * The code is partially based on the BinaryThin2 plugin by G. Landini
 * (http://www.dentistry.bham.ac.uk/landinig/software/software.html).
 * The HitOrMiss function is called using the Homotopic Marking
 * parameters:
 *   kernel_a=[2 1 2 0 1 0 0 0 0 ]
 *   kernel_b=[0 0 0 1 1 1 1 1 1 ]
 *   rotations=[rotate 45]
 *   iterations=-1
 *  
 */

import java.awt.image.IndexColorModel;

import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;

public class Skeleton_Intersections implements PlugInFilter {
	
	private boolean whiteBackground;
	private boolean showResultImage;
	private boolean showResultsTable;
	private int numOfIntersections; 
	private String originalImageTitle;

	public int setup( String arg, ImagePlus imp ) {
		
		ImageStatistics stats;
		stats = imp.getStatistics();
		if ( stats.histogram[0] + stats.histogram[255] != stats.pixelCount ){
			IJ.error( "8-bit binary image (0 and 255) required." );
			return DONE;
		}
		
		originalImageTitle = imp.getTitle();

		GenericDialog gd = new GenericDialog( "Skeleton Intersections", IJ.getInstance() );
		gd.addMessage( "Skeleton Intersections" );
		gd.addCheckbox( "White background", false );
		gd.addCheckbox( "Pseudo colored result image", false );
		gd.addCheckbox( "Results table", false );

		gd.showDialog();
		if ( gd.wasCanceled() ) {
			return DONE;
		}

		whiteBackground = gd.getNextBoolean();
		showResultImage = gd.getNextBoolean();
		showResultsTable = gd.getNextBoolean();
		
		return DOES_8G + DOES_STACKS;
		
	}

	public void run( ImageProcessor ip ) {

		ImageProcessor originalIP = ip.duplicate();
		numOfIntersections = 0;
		
		int xe = ip.getWidth();
		int ye = ip.getHeight();

		int [][] pixels1 = new int [xe][ye];
		int [][] pixels2 = new int [xe][ye];		
		
		if ( whiteBackground ) {
			ip.invert();
		}		

		// get the original pixels
		for ( int y = 0; y < ye; y++ ) {
			for ( int x = 0; x < xe; x++ ) {
				pixels1[x][y] = ip.getPixel( x, y );
				pixels2[x][y] = 0;
			}
		}

		// detect the intersecting pixels
		int localHits = 0;
		for ( int y = 1; y < (ye-1); y++ ) {
			for ( int x = 1; x < (xe-1); x++ ) {
				if ( pixels1[x][y] == 255 ) {
					// a 3x3 neighborhood
					if ( pixels1[ x-1 ][ y-1 ] == 255 ) { localHits++; }
					if ( pixels1[ x   ][ y-1 ] == 255 ) { localHits++; }
					if ( pixels1[ x+1 ][ y-1 ] == 255 ) { localHits++; }
					if ( pixels1[ x-1 ][ y   ] == 255 ) { localHits++; }
					if ( pixels1[ x+1 ][ y   ] == 255 ) { localHits++; }
					if ( pixels1[ x-1 ][ y+1 ] == 255 ) { localHits++; }
					if ( pixels1[ x   ][ y+1 ] == 255 ) { localHits++; }
					if ( pixels1[ x+1 ][ y+1 ] == 255 ) { localHits++; }
					if ( localHits > 2 ) {
						// we have an intersection, mark it with a 1
						pixels2[x][y] = 1;
					}
					localHits = 0;
				}
			}		
		}

		int []lut = new int [512];
		int[] kA = {2, 1, 2, 0, 1, 0, 0, 0, 0}; 
		int[] kB = {0, 0, 0, 1, 1, 1, 1, 1, 1};	

		for ( int y = 0; y < ye; y++ ) {
			for ( int x = 0; x < xe; x++ ){
				pixels1[x][y] = 0;
			}
		}

		int currCount = 1; // loop at least once
		while ( currCount > 0 ) {

			// do it the first time
			createHMlut( kA, lut );
			hitOrMiss( xe, ye, pixels2, pixels1, lut );
			createHMlut( kB, lut );
			hitOrMiss( xe, ye, pixels2, pixels1, lut );

			// "rotate 45", i.e. rotate left
			for( int i = 0; i < 7; i++ ) {
				rotate45( kA );
				createHMlut( kA, lut );
				hitOrMiss( xe, ye, pixels2, pixels1, lut );
				rotate45( kB );
				createHMlut( kB, lut );
				hitOrMiss( xe, ye, pixels2, pixels1, lut );
			}
			// put back in order for next iteration
			rotate45( kA );
			rotate45( kB );

			// keep looping until we have no positive pixels
			currCount = 0;
			for ( int x = 0; x < xe; x++ ) {
				for ( int y = 0; y < ye; y++ ) {
					currCount += pixels1[x][y];
					pixels1[x][y] = 0;
				}
			}
		}	
		
		// calculate the number of found intersections
		// (i.e., # of white pixels in the image)
		for ( int x = 0; x < xe; x++ ) {
			for (int y = 0 ; y < ye ; y++ ){
				if ( pixels2[x][y] > 0 ) {
					numOfIntersections++;
				}
			}
		}

		// clear the original image
		ip.setValue( 0 );
		ip.fill();		

		// update the original image to correspond to the result
		for ( int x = 0; x < xe; x++ ) {
			for ( int y = 0; y < ye; y++ ){
				ip.putPixel( x, y, (pixels2[x][y] > 0 ? 255:0 ) );
			}
		}
		
		if ( showResultImage ) {

			ImageProcessor cip = ip.duplicate();
			cip.erode();
	        IndexColorModel icm = (IndexColorModel) ip.getColorModel();
	        int mapSize = icm.getMapSize();
	        byte[] reds = new byte[mapSize];
	        byte[] greens = new byte[mapSize];
	        byte[] blues = new byte[mapSize];   
	        icm.getReds(reds); 
	        icm.getGreens(greens); 
	        icm.getBlues(blues);
	        greens[255] = 0;
	        blues[255] = 0;
	        IndexColorModel newIcm = new IndexColorModel( 8, 256, reds, greens, blues );	        
	        cip.setColorModel( newIcm );

			ImagePlus resultImage = new ImagePlus( "Skeleton intersections", originalIP.convertToRGB() );
	        resultImage.getProcessor().copyBits(
					cip, 0, 0, Blitter.COPY_ZERO_TRANSPARENT );
			resultImage.updateImage();
			resultImage.show();
			
		}
		
		if ( showResultsTable ) {
			IJ.setColumnHeadings( "Image name	Num of intersections" );
			IJ.getTextPanel().append( originalImageTitle + "	" + String.valueOf( numOfIntersections ) );
		}			

		if ( whiteBackground ) {
			ip.invert();
		}

	}
	
	public int getNumOfIntersections() {
		return numOfIntersections;		
	}
	
	void rotate45 (int [] kernel){
		int temp;
		temp=kernel[0];
		kernel[0]=kernel[1];
		kernel[1]=kernel[2];
		kernel[2]=kernel[5];
		kernel[5]=kernel[8];
		kernel[8]=kernel[7];
		kernel[7]=kernel[6];
		kernel[6]=kernel[3];
		kernel[3]=temp;
	}

	void createHMlut(int [] kernel, int [] lut){
		int i, j, match, toMatch;

		for(i=0;i<512;i++)
			lut[i]=1;

		toMatch=0;
		for(j=0;j<9;j++){
			if (kernel[j]!=2)
				toMatch++;
		}
		//System.out.println("Debug: to match: "+toMatch);

		//make lut
		for(i=0;i<512;i++){
			match=0;
			for(j=0;j<9;j++){
				if (kernel[j]!=2)
					if ((((i & (int)Math.pow(2,j))!=0)?1:0) == kernel[j]) match++;
			}
			if (match!=toMatch){
				//System.out.println("Don't match:"+i);
				lut[i]=0;
			}
		}
	}

	void hitOrMiss(int xe, int ye, int [][] pixel2, int [][]pixel, int [] lut){
		int x, y;
		// considers the pixels outside image to be 0 (not set)
		for (x=1; x<xe-1; x++) {
			for (y=1; y<ye-1; y++) {
				pixel[x][y]+= lut[(
				pixel2[x-1][y-1] +
				pixel2[x  ][y-1] * 2 +
				pixel2[x+1][y-1] * 4 +
				pixel2[x-1][y  ] * 8 +
				pixel2[x  ][y  ] * 16 +
				pixel2[x+1][y  ] * 32 +
				pixel2[x-1][y+1] * 64 +
				pixel2[x  ][y+1] * 128 +
				pixel2[x+1][y+1] * 256)];
			}
		}

		y=0;
		for (x=1; x<xe-1; x++) {
			//upper row
			pixel[x][y]+= lut[(
			pixel2[x-1][y  ] * 8 +
			pixel2[x  ][y  ] * 16 +
			pixel2[x+1][y  ] * 32 +
			pixel2[x-1][y+1] * 64 +
			pixel2[x  ][y+1] * 128 +
			pixel2[x+1][y+1] * 256)];
		}

		y=ye-1;
		for (x=1; x<xe-1; x++) {
			//lower row
			pixel[x][y]+= lut[(
			pixel2[x-1][y-1] +
			pixel2[x  ][y-1] * 2 +
			pixel2[x+1][y-1] * 4 +
			pixel2[x-1][y  ] * 8 +
			pixel2[x  ][y  ] * 16 +
			pixel2[x+1][y  ] * 32)];
		}

		x=0;
		for (y=1; y<ye-1; y++) {
			//left column
			pixel[x][y]+= lut[(
			pixel2[x  ][y-1] * 2 +
			pixel2[x+1][y-1] * 4 +
			pixel2[x  ][y  ] * 16 +
			pixel2[x+1][y  ] * 32 +
			pixel2[x  ][y+1] * 128 +
			pixel2[x+1][y+1] * 256)];
		}

		x=xe-1;
		for (y=1; y<ye-1; y++) {
			//right column
			pixel[x][y]+= lut[(
			pixel2[x-1][y-1] +
			pixel2[x  ][y-1] * 2 +
			pixel2[x-1][y  ] * 8 +
			pixel2[x  ][y  ] * 16 +
			pixel2[x-1][y+1] * 64 +
			pixel2[x  ][y+1] * 128)];
		}

		x=0; //upper left corner
		y=0;
		pixel[x][y]+= lut[(pixel2[x][y] * 16 + pixel2[x+1][y] * 32 + pixel2[x][y+1] * 128 + pixel2[x+1][y+1] * 256)];

		x=xe-1; //upper right corner
		//y=0;
		pixel[x][y]+= lut[(pixel2[x-1][y] * 8 + pixel2[x][y] * 16 + pixel2[x-1][y+1] * 64 + pixel2[x][y+1] * 128)];

		x=0; //lower left corner
		y=ye-1;
		pixel[x][y]+= lut[(pixel2[x][y-1] * 2 + pixel2[x+1][y-1] * 4 + pixel2[x][y] * 16 + pixel2[x+1][y] * 32)];

		x=xe-1; //lower right corner
		y=ye-1;
		pixel[x][y]+= lut[(pixel2[x-1][y-1] + pixel2[x][y-1] * 2 + pixel2[x-1][y] * 8 + pixel2[x  ][y] * 16)];

		for (x=0; x<xe; x++) {
			for (y=0; y<ye; y++){
				pixel2[x][y]= ((pixel2[x][y]-pixel[x][y])>0?1:0);
				//pixel2[x][y]-= pixel[x][y];
			}
		}
	}

}
